from json import JSONDecodeError
from typing import Any

import jmespath
import requests
from django import forms
from jmespath.exceptions import LexerError

from ..exceptions import CheckError
from . import HttpCheck
from .http import BaseHttpConfig


class JsonConfig(BaseHttpConfig):
    match = forms.CharField(required=False)


class JsonCheck(HttpCheck):
    icon = "json.svg"
    pragma = ["http+json", "https+json"]
    config_class = JsonConfig

    @classmethod
    def clean_config(cls, cfg: dict[str, Any]) -> dict[str, Any]:
        if not cfg.get("url"):
            cfg["url"] = cfg.get("address", "").replace("+json://", "://")
        return cfg

    def check(self, raise_error: bool = False) -> bool:
        try:
            timeout = self.config["timeout"]
            match = self.config["match"]
            res = requests.get(self.config["url"], timeout=timeout)
            if res.status_code not in self.config["status_success"]:
                return False
            data = res.json()
            if match:
                return jmespath.search(match, data)
            return True
        except (forms.ValidationError, requests.exceptions.RequestException, JSONDecodeError, LexerError) as e:
            if raise_error:
                raise CheckError("JSON check failed") from e
        return False



from pathlib import Path

from admin_extra_buttons.decorators import button
from admin_extra_buttons.mixins import ExtraButtonsMixin
from adminfilters.autocomplete import AutoCompleteFilter, LinkedAutoCompleteFilter
from adminfilters.mixin import AdminFiltersMixin
from django.conf import settings
from django.contrib import admin, messages
from django.contrib.auth.admin import UserAdmin as _UserAdmin
from django.db.models import Model, QuerySet
from django.http import Http404, HttpRequest, HttpResponse, HttpResponseRedirect
from django.shortcuts import render
from django.templatetags.static import static
from django.urls import reverse
from flags.models import FlagState
from unfold.admin import ModelAdmin as UnfoldModelAdmin

from .forms import ChangeIconForm, FlagStateForm, MonitorForm
from .models import Deadline, Environment, LogCheck, Monitor, Project, User
from .tasks import queue_trigger
from .ws.utils import notify_ui


class BirderAdminMixin(ExtraButtonsMixin, AdminFiltersMixin, UnfoldModelAdmin):
    pass


@admin.register(User)
class UserAdmin(_UserAdmin[User]):
    search_fields = ("username",)


@admin.register(Project)
class ProjectAdmin(BirderAdminMixin, admin.ModelAdmin[Project]):
    search_fields = ("name",)
    list_display = (
        "name",
        "public",
    )
    filter_horizontal = ("environments",)

    def get_changeform_initial_data(self, request: HttpRequest) -> dict:
        initial = super().get_changeform_initial_data(request)
        initial.setdefault("environments", (Environment.objects.first()))
        return initial

    @button()
    def monitors(self, request: HttpRequest, pk: str) -> HttpResponseRedirect:
        url = reverse("admin:birder_monitor_changelist")
        return HttpResponseRedirect(f"{url}?project__exact={pk}")


def assert_object_or_404(obj: Model | None) -> None:
    if not obj:
        raise Http404


@admin.register(Monitor)
class MonitorAdmin(BirderAdminMixin, admin.ModelAdmin[Monitor]):
    search_fields = ("name",)
    list_display = ("name", "project", "environment", "counters", "checker", "active")
    list_filter = (
        ("project", LinkedAutoCompleteFilter.factory(parent=None)),
        ("environment", LinkedAutoCompleteFilter.factory(parent=None)),
        "strategy",
        "active",
    )
    actions = ["check_selected"]
    autocomplete_fields = ("environment", "project")
    form = MonitorForm
    change_form_template = None

    @admin.display(ordering="strategy")
    def checker(self, obj: Monitor) -> bool:
        return obj.strategy.__class__.__name__

    @admin.display()
    def counters(self, obj: Monitor) -> str:
        return "{} / {} / {}".format(*obj.counters)

    def check_selected(self, request: HttpRequest, queryset: QuerySet[Monitor]) -> None:
        for m in queryset.all():
            queue_trigger.send(m.id)

    def get_fields(self, request: HttpRequest, obj: Monitor | None = None) -> list[str]:
        return [
            "name",
            "strategy",
            "project",
            "environment",
            "active",
            "warn_threshold",
            "err_threshold",
            "description",
            "notes",
        ]

    @button(label="Refresh Token")
    def regenerate_token(self, request: HttpRequest, pk: str) -> HttpResponse:
        self.get_common_context(request, pk)
        self.object.regenerate_token(True)

    @button(label="Change Icon")
    def change_icon(self, request: HttpRequest, pk: str) -> HttpResponse:
        ctx = self.get_common_context(request, pk)
        assert_object_or_404(self.object)
        ctx["icons"] = sorted(
            [
                (p.name, request.build_absolute_uri(static(f"images/icons/{p.name}")))
                for p in (Path(settings.PACKAGE_DIR) / "static" / "images" / "icons").glob("*.*")
            ]
        )
        if request.method == "POST":
            form = ChangeIconForm(request.POST)
            if form.is_valid():
                self.object.custom_icon = form.cleaned_data["icon"]
                self.object.save()
                notify_ui("refresh", monitor=self.object, crud="update")
                return HttpResponseRedirect("..")
        form = ChangeIconForm(initial={"icon": self.object.custom_icon})
        ctx["form"] = form

        return render(request, "admin/birder/monitor/change_icon.html", ctx)

    @button(label="Check")
    def manual_check(self, request: HttpRequest, pk: str) -> HttpResponse:
        self.get_common_context(request, pk)
        monitor: Monitor = self.object
        assert_object_or_404(monitor)

        if monitor.run():
            self.message_user(request, "Monitor checked", level=messages.SUCCESS)
        else:
            self.message_user(request, "Monitor failed", level=messages.ERROR)

    @button()
    def configure(self, request: HttpRequest, pk: str) -> HttpResponse:
        ctx = self.get_common_context(request, pk)
        monitor: Monitor = self.object
        assert_object_or_404(monitor)
        if monitor.strategy.config_class:
            if request.method == "POST":
                form = monitor.strategy.config_class(request.POST, initial=monitor.configuration)
                if form.is_valid():
                    monitor.configuration = form.cleaned_data
                    if "check" in request.POST:
                        if not monitor.strategy.check():
                            self.message_user(request, "Check failed", level=messages.ERROR)
                        else:
                            self.message_user(request, "Check success", level=messages.SUCCESS)
                            monitor.save()
                            return HttpResponseRedirect("..")
                    else:
                        monitor.save()
                        return HttpResponseRedirect("..")
            else:
                form = monitor.strategy.config_class(initial=monitor.configuration)
            ctx["form"] = form
            ctx["form_help"] = form.render_help(ctx, request=request, monitor=self.object)
        return render(request, "admin/birder/monitor/configure.html", ctx)


@admin.register(LogCheck)
class LogCheckAdmin(BirderAdminMixin, admin.ModelAdmin[LogCheck]):
    list_display = ("timestamp", "monitor", "status")
    list_filter = ("status", "timestamp", ("monitor", AutoCompleteFilter))
    readonly_fields = ("timestamp", "monitor", "status", "payload")

    def has_add_permission(self, request: HttpRequest) -> bool:
        return False


@admin.register(Environment)
class EnvironmentAdmin(BirderAdminMixin, admin.ModelAdmin[LogCheck]):
    list_display = ("name",)
    search_fields = ("name",)


@admin.register(Deadline)
class DeadlineAdmin(ExtraButtonsMixin, AdminFiltersMixin, admin.ModelAdmin[LogCheck]):
    autocomplete_fields = ("monitor",)

    list_display = ("monitor", "start", "end", "recurrences")
    search_fields = ("monitor__name",)


admin.site.unregister(FlagState)


@admin.register(FlagState)
class FlagStateAdmin(BirderAdminMixin, admin.ModelAdmin[FlagState]):
    form = FlagStateForm



from typing import Any

import click
from django.core.management import call_command

from . import cli


@cli.command()
@click.option("--deploy", is_flag=True, default=False)
@click.option("--list-tags", is_flag=True, default=False)
@click.option("-v", "--verbosity", type=int, default=1)
def check(deploy: bool, verbosity: int, **kwargs: Any) -> None:
    """Check Birder configuration."""
    call_command("check", deploy=deploy, verbosity=verbosity, **kwargs)



from typing import Any

from django.conf import settings
from django.http import HttpRequest
from django.templatetags.static import static
from django.urls import reverse_lazy

COMMON_CONFIG = {
    "SITE_TITLE": "Birder: ",
    "SITE_HEADER": "Birder",
    "SITE_SUBHEADER": "Appears under SITE_HEADER",
    "SITE_FAVICONS": [
        {
            "rel": "icon",
            "sizes": "32x32",
            "type": "image/x-icon",
            "href": lambda request: static("itcaster/images/favicon.ico"),
        },
    ],
    "SITE_URL": "/",
    "SITE_ICON": {
        "light": lambda request: static("images/birder.svg"),  # light mode
        "dark": lambda request: static("images/birder_dark.svg"),  # dark mode
    },
    "SITE_LOGO": {
        "light": lambda request: static("images/birder.svg"),  # light mode
        "dark": lambda request: static("images/birder_dark.svg"),  # dark mode
    },
    "STYLES": [
        lambda request: static("/css/styles_admin.css"),
    ],
    "BORDER_RADIUS": "6px",
    "COLORS": {
        "base": {
            "50": "249, 250, 251",
            "100": "243, 244, 246",
            "200": "229, 231, 235",
            "300": "209, 213, 219",
            "400": "156, 163, 175",
            "500": "107, 114, 128",
            "600": "75, 85, 99",
            "700": "55, 65, 81",
            "800": "31, 41, 55",
            "900": "17, 24, 39",
            "950": "3, 7, 18",
        },
        "primary": {
            "50": "254, 242, 242",
            "100": "254, 226, 226",
            "200": "254, 202, 202",
            "300": "252, 165, 165",
            "400": "248, 113, 113",
            "500": "239, 68, 68",
            "600": "220, 38, 38",
            "700": "185, 28, 28",
            "800": "153, 27, 27",
            "900": "127, 29, 29",
            "950": "76, 23, 23",
        },
        "font": {
            "subtle-light": "var(--color-base-500)",  # text-base-500
            "subtle-dark": "var(--color-base-400)",  # text-base-400
            "default-light": "var(--color-base-600)",  # text-base-600
            "default-dark": "var(--color-base-300)",  # text-base-300
            "important-light": "var(--color-base-900)",  # text-base-900
            "important-dark": "var(--color-base-100)",  # text-base-100
        },
    },
}
UNFOLD = {
    **COMMON_CONFIG,
    "ENVIRONMENT": "birder.config.fragments.unfold.environment_callback",  # environment name in header
    "SHOW_VIEW_ON_SITE": True,  # show/hide "View on site" button, default: True
    "LOGIN": {
        "image": lambda request: static("images/birder.svg"),
        "redirect_after": lambda request: reverse_lazy("admin:index"),
    },
    "SIDEBAR": {
        "show_search": True,  # Search in applications and models names
        "show_all_applications": True,  # Dropdown with all applications and models
    },
}
MANAGE_CONFIG = {
    **COMMON_CONFIG,
    "SITE_SYMBOL": "speed",  # symbol from icon set
    "SHOW_HISTORY": True,  # show/hide "History" button, default: True
    "SHOW_BACK_BUTTON": False,  # show/hide "Back" button on changeform in header, default: False
    "THEME": "dark",  # Force theme: "dark" or "light". Will disable theme switcher
    "LOGIN": {
        "image": lambda request: static("images/birder.svg"),
        "redirect_after": lambda request: reverse_lazy("manage:index"),
    },
}


def dashboard_callback(request: HttpRequest, context: dict[str, Any]) -> dict[str, Any]:
    context.update(
        {
            "sample": "example",  # this will be injected into templates/admin/index.html
        }
    )
    return context


def environment_callback(request: "HttpRequest") -> tuple[str, str]:
    return settings.ENVIRONMENT


def badge_callback(request: "HttpRequest") -> int:
    return 3


def permission_callback(request: "HttpRequest") -> bool:
    return request.user.is_superuser



CONSTANCE_BACKEND = "constance.backends.database.DatabaseBackend"


CONSTANCE_ADDITIONAL_FIELDS = {
    "email": [
        "django.forms.EmailField",
        {},
    ],
    "group_select": [
        "birder.utils.constance.GroupChoiceField",
        {"initial": None},
    ],
    "duration": [
        "birder.utils.constance.DurationField",
        {"initial": None},
    ],
    "read_only_text": [
        "django.forms.fields.CharField",
        {
            "required": False,
            "widget": "birder.utils.constance.ObfuscatedInput",
        },
    ],
    "write_only_text": [
        "django.forms.fields.CharField",
        {
            "required": False,
            "widget": "birder.utils.constance.WriteOnlyTextarea",
        },
    ],
    "write_only_input": [
        "django.forms.fields.CharField",
        {
            "required": False,
            "widget": "birder.utils.constance.WriteOnlyInput",
        },
    ],
}

CONSTANCE_CONFIG = {
    "NEW_USER_IS_STAFF": (False, "Set NEW_USER_DEFAULT_GROUP new user as staff", bool),
    "NEW_USER_DEFAULT_GROUP": (
        None,
        "Group to assign to any new user",
        "group_select",
    ),
    "HARD_THRESHOLD": (86400, "System Wide Threshold", int),
    "RETENTION_POLICY": ("1 month", "System Wide Threshold", "duration"),
    "CACHE_PREFIX": ("birder", "Cache global prefix", str),
    "DATETIME_FORMAT": ("%Y %b %d %H:%M", "Datetime format", str),
    "DATE_FORMAT": ("%Y %b %d", "Datetime format", str),
    "TIME_FORMAT": ("%H:%M", "Time format", str),
    "TOKEN_LENGTH": (255, "Monitor token length", int),
}



from typing import Any

from django import forms
from django.contrib.auth.forms import AuthenticationForm
from flags.forms import FlagStateForm as FlagStateForm_
from unfold.widgets import BASE_INPUT_CLASSES, CHECKBOX_CLASSES, SELECT_CLASSES

from birder.models import Monitor, Project, User


class DateInput(forms.DateInput):
    input_type = "date"


class LoginForm(AuthenticationForm):
    username = forms.EmailField(label="Email", widget=forms.TextInput(attrs={"autofocus": True}))

    class Meta:
        model = User
        fields = ("username", "password")


class ChangeIconForm(forms.Form):
    icon = forms.URLField(required=False)

    @property
    def media(self) -> forms.Media:
        media = super().media
        media += forms.Media(
            js=[
                "admin/js/vendor/jquery/jquery.js",
                "admin/js/jquery.init.js",
                "change-icon.js",
            ],
            css={"screen": ["birder-admin.css"]},
        )
        return media


class MonitorForm(forms.ModelForm):
    class Meta:
        model = Monitor
        fields = (
            "project",
            "environment",
            "name",
            "position",
            "description",
            "notes",
            "strategy",
            "configuration",
        )


class ProjectForm(forms.ModelForm):
    class Meta:
        model = Project
        fields = ("name", "environments", "public", "bitcaster_url", "icon")


class FlagStateForm(FlagStateForm_):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        super().__init__(*args, **kwargs)

        self.fields["name"].widget.attrs["class"] = " ".join(SELECT_CLASSES)
        self.fields["condition"].widget.attrs["class"] = " ".join(SELECT_CLASSES)
        self.fields["value"].widget.attrs["class"] = " ".join(BASE_INPUT_CLASSES)
        self.fields["required"].widget.attrs["class"] = " ".join(CHECKBOX_CLASSES)



import os
from pathlib import Path

from . import env

# Build paths inside the project like this: BASE_DIR / 'subdir'.
PACKAGE_DIR = Path(__file__).resolve().parent.parent  # src/birder
PROJECT_DIR = PACKAGE_DIR.parent.parent  # git project root

# Quick-start development settings - unsuitable for production
# See https://docs.djangoproject.com/en/5.1/howto/deployment/checklist/

# SECURITY WARNING: keep the secret key used in production secret!
SECRET_KEY = env("SECRET_KEY")

# SECURITY WARNING: don't run with debug turned on in production!
DEBUG = env("DEBUG")

ALLOWED_HOSTS = env("ALLOWED_HOSTS")

# Application definition

INSTALLED_APPS = [
    "birder.theme",
    "daphne",
    "unfold",  # before django.contrib.admin
    "unfold.contrib.filters",  # optional, if special filters are needed
    "unfold.contrib.forms",  # optional, if special form elements are needed
    "unfold.contrib.inlines",  # optional, if special inlines are needed
    "unfold.contrib.simple_history",  # optional, if dja
    "django.contrib.admin",
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.messages",
    "django.contrib.staticfiles",
    "admin_extra_buttons",
    "markdown_deux",
    "django_dramatiq",
    "dramatiq_crontab",
    "adminfilters",
    "constance",
    "recurrence",
    "flags",
    "social_django",
    "tailwind",
    "birder",
    "birder.ws",
    *env("EXTRA_APPS"),
]

MIDDLEWARE = [
    "django.middleware.security.SecurityMiddleware",
    "django.contrib.sessions.middleware.SessionMiddleware",
    "django.middleware.common.CommonMiddleware",
    "django.middleware.csrf.CsrfViewMiddleware",
    "django.contrib.auth.middleware.AuthenticationMiddleware",
    "django.contrib.messages.middleware.MessageMiddleware",
    "django.middleware.clickjacking.XFrameOptionsMiddleware",
]
SESSION_ENGINE = "django.contrib.sessions.backends.signed_cookies"
SESSION_EXPIRE_AT_BROWSER_CLOSE = False
SESSION_COOKIE_NAME = "birder_sessionid"

ROOT_URLCONF = "birder.config.urls"

TEMPLATES = [
    {
        "BACKEND": "django.template.backends.django.DjangoTemplates",
        "DIRS": [],
        "APP_DIRS": True,
        "OPTIONS": {
            "context_processors": [
                "django.template.context_processors.debug",
                "django.template.context_processors.request",
                "django.contrib.auth.context_processors.auth",
                "django.contrib.messages.context_processors.messages",
                "birder.context_processors.birder",
            ],
        },
    },
]

WSGI_APPLICATION = "birder.config.wsgi.application"

ASGI_APPLICATION = "birder.config.asgi.application"
CHANNEL_LAYERS = {
    "default": {
        "BACKEND": "channels_redis.core.RedisChannelLayer",
        "CONFIG": {
            "hosts": [env("VALKEY_URL")],
        },
    },
}

# Database
# https://docs.djangoproject.com/en/5.1/ref/settings/#databases

DATABASES = {"default": env.db("DATABASE_URL")}
CACHE_URL = env("CACHE_URL")
CACHES = {
    "default": {
        "BACKEND": "redis_lock.django_cache.RedisCache",
        "LOCATION": CACHE_URL,
        "OPTIONS": {"CLIENT_CLASS": "django_redis.client.DefaultClient"},
    }
}

# Password validation
# https://docs.djangoproject.com/en/5.1/ref/settings/#auth-password-validators
LOGIN_URL = "/login/"
LOGIN_REDIRECT_URL = "/"
LOGOUT_REDIRECT_URL = "/"

AUTH_USER_MODEL = "birder.User"

AUTH_PASSWORD_VALIDATORS = [
    {
        "NAME": "django.contrib.auth.password_validation.UserAttributeSimilarityValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.MinimumLengthValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.CommonPasswordValidator",
    },
    {
        "NAME": "django.contrib.auth.password_validation.NumericPasswordValidator",
    },
]

AUTHENTICATION_BACKENDS = (
    "social_core.backends.google.GoogleOAuth2",
    "django.contrib.auth.backends.ModelBackend",
)

# Internationalization
# https://docs.djangoproject.com/en/5.1/topics/i18n/

LANGUAGE_CODE = "en-us"

TIME_ZONE = "UTC"

USE_I18N = True

USE_TZ = True

# Static files (CSS, JavaScript, Images)
# https://docs.djangoproject.com/en/5.1/howto/static-files/

STATIC_URL = env("STATIC_URL")
STATIC_ROOT = env("STATIC_ROOT")

# Default primary key field type
# https://docs.djangoproject.com/en/5.1/ref/settings/#default-auto-field

DEFAULT_AUTO_FIELD = "django.db.models.BigAutoField"


def get_log_level_for(app: str) -> str:
    return os.environ.get(f"{app.upper()}_LOG_LEVEL", env("LOG_LEVEL"))


LOGGING = {
    "version": 1,
    "disable_existing_loggers": False,
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
        },
    },
    "root": {
        "handlers": ["console"],
        "level": "WARNING",
    },
    "loggers": {
        "django": {
            "handlers": ["console"],
            "level": get_log_level_for("django"),
            "propagate": False,
        },
        "birder": {
            "handlers": ["console"],
            "level": get_log_level_for("birder"),
            "propagate": False,
        },
    },
}
CSRF_TRUSTED_ORIGINS = env("CSRF_TRUSTED_ORIGINS")
SECURE_PROXY_SSL_HEADER = ("HTTP_X_FORWARDED_PROTO", "https")
SECURE_HSTS_SECONDS = env("SECURE_HSTS_SECONDS")

from .fragments.app import *  # noqa
from .fragments.constance import *  # noqa
from .fragments.crypt import *  # noqa

from .fragments.dramatiq import *  # noqa
from .fragments.social_auth import *  # noqa
from .fragments.tailwind import *  # noqa
from .fragments.unfold import *  # noqa



import os
from typing import Any

import click
from click import Context

from . import cli

KEY = "birder:upgrade"


@cli.command()
@click.option("--force/--no-force", is_flag=True, default=False)
@click.option("--check/--no-check", is_flag=True, default=False)
@click.option("--clear/--no-clear", is_flag=True, default=False)
@click.option("-v", "--verbosity", type=int, default=1)
@click.pass_context
def upgrade(ctx: Context, force: bool, verbosity: int, check: bool, clear: bool, **kwargs: Any) -> None:  # noqa: C901, PLR0912
    """Upgrade system."""
    from constance import config
    from django.contrib.auth.models import Group
    from django.core.cache import cache
    from django.core.management import call_command

    from birder.models import Environment, User

    redis_client = cache.client.get_client()
    if force or clear:
        redis_client.delete(KEY)
    if clear:
        return
    if check:
        if redis_client.exists(KEY):
            click.secho("Upgrade is running.", fg="yellow")
        else:
            click.secho("Upgrade is not running.", fg="green")
        return
    if redis_client.set(KEY, "locked", nx=True, ex=86400):
        try:
            if verbosity >= 1:
                click.secho("Run database migrations")
            call_command("migrate", interactive=False, verbosity=verbosity - 1)
            if verbosity >= 1:
                click.secho("Collect static assets")
            call_command("collectstatic", interactive=False, verbosity=-1)
            if verbosity >= 1:
                click.secho("Create standard environments")
            Environment.objects.get_or_create(name="development")
            Environment.objects.get_or_create(name="staging")
            Environment.objects.get_or_create(name="production")

            g, is_new = Group.objects.get_or_create(name="Default")
            if is_new:
                config.NEW_USER_DEFAULT_GROUP = g.pk
            if (admin_user_email := os.environ.get("ADMIN_EMAIL")) and os.environ.get("ADMIN_PASSWORD"):
                try:
                    User.objects.get(email=admin_user_email)
                    click.secho(f"{admin_user_email} user found. Superuser not updated/created", fg="yellow")
                except User.DoesNotExist:
                    User.objects.create_superuser(
                        username=admin_user_email, email=admin_user_email, password=os.environ.get("ADMIN_PASSWORD")
                    )
                    click.secho("Superuser created!", fg="green")
            elif verbosity >= 1:
                click.secho("no ADMIN_EMAIL/ADMIN_PASSWORD env vars found", fg="yellow")
        except Exception as e:  # noqa: BLE001
            redis_client.delete(KEY)
            click.secho(f"{e}", fg="red", err=True)
            raise click.Abort(2) from None
        finally:
            redis_client.delete(KEY)
    else:
        click.secho("Concurrent process detected.", fg="red", err=True)
        ctx.exit(2)



from datetime import datetime
from typing import Any

from django.db.models.base import Model
from django.db.models.signals import post_delete, post_save
from django.dispatch import receiver

from birder.models import Monitor
from birder.signals import monitor_update
from birder.ws.utils import notify_ui


# method for updating
@receiver(post_save, sender=Monitor, dispatch_uid="update_monitor")
def update_monitor(sender: type[Model], instance: Monitor, created: bool, **kwargs: Any) -> None:
    if created:
        notify_ui("refresh", monitor=instance, crud="add")


# method for updating
@receiver(post_delete, sender=Monitor, dispatch_uid="delete_monitor")
def delete_monitor(sender: type[Model], instance: Monitor, **kwargs: Any) -> None:
    notify_ui("refresh", monitor=instance, crud="delete")


# method for updating
@receiver(monitor_update, sender=Monitor, dispatch_uid="monitor_check")
def monitor_check(sender: type[Model], instance: Monitor, result: bool, timestamp: datetime, **kwargs: Any) -> None:
    pass



import kombu.exceptions
import redis.exceptions
from django import forms
from django.core.validators import MaxValueValidator, MinValueValidator
from redis import Redis as RedisClient

from ..exceptions import CheckError
from .base import BaseCheck, ConfigForm, WriteOnlyField


class RedisConfig(ConfigForm):
    host = forms.CharField(required=True, help_text="Server hostname or IP Address")
    port = forms.IntegerField(validators=[MinValueValidator(1)], initial=6379)
    socket_timeout = forms.IntegerField(validators=[MinValueValidator(1), MaxValueValidator(5)], initial=2)
    socket_connect_timeout = forms.IntegerField(validators=[MinValueValidator(1), MaxValueValidator(5)], initial=2)
    password = WriteOnlyField(required=False)


class RedisCheck(BaseCheck):
    icon = "redis.svg"
    pragma = ["redis"]
    config_class = RedisConfig
    address_format = "{host}:{port}"

    def check(self, raise_error: bool = False) -> bool:
        try:
            client = RedisClient(**self.config)
            client.ping()
            return True
        except (redis.exceptions.ConnectionError, ConnectionRefusedError, kombu.exceptions.KombuError) as e:
            if raise_error:
                raise CheckError("Redis check failed") from e
        return False



from django.conf import settings
from django.contrib import admin
from django.contrib.auth.views import LogoutView
from django.urls import include, path

from birder import views
from birder.manage.site import ManageSite

handler400 = views.error_400
handler403 = views.error_403
handler404 = views.error_404
handler500 = views.error_500

manage = ManageSite("manage")
urlpatterns = [
    path("errors/400/", handler400, name="errors-400"),
    path("errors/403/", handler403, name="errors-403"),
    path("errors/404/", handler404, name="errors-404"),
    path("errors/500/", handler500, name="errors-500"),
    path("social/", include("social_django.urls", namespace="social")),
    path("admin/", admin.site.urls),
    path("manage/", manage.urls),
    path("login/", views.LoginView.as_view(), name="login"),
    path("logout/", LogoutView.as_view(), name="logout"),
    path("trigger/<int:pk>/<str:token>/", views.trigger, name="trigger"),
    path("<int:project_id>/<str:env>/<int:pk>", views.MonitorDetail.as_view(), name="monitor-detail"),
    path("<int:project_id>/<str:env>/", views.ProjectView.as_view(), name="project-env"),
    path("<int:pk>/", views.ProjectRouterView.as_view(), name="project-detail"),
    path("", views.IndexView.as_view(), name="index"),
]
if "django_browser_reload" in settings.INSTALLED_APPS:  # pragma: no cover
    urlpatterns += [path(r"__reload__/", include("django_browser_reload.urls"))]


admin.autodiscover()
admin.site.site_header = "Birder"
admin.site.site_title = "Birder site admin"
admin.site.index_title = ""



from typing import TYPE_CHECKING, Any

from django.conf import settings
from django.core import checks

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence

    from django.apps import AppConfig


def check_crypt(
    app_configs: "Sequence[AppConfig]", databases: "Sequence[str] | None", **kwargs: Any
) -> "Iterable[checks.CheckMessage]":
    errors = []
    if not settings.SALT_KEY:
        errors.append(checks.Error("CRYPT_SALT_KEY env var must be a list of Fernet keys"))
    if not settings.SECRET_KEY_FALLBACKS:
        errors.append(checks.Error("CRYPT_KEYS env var must be a list of Fernet keys"))
    return errors



from collections.abc import Iterable
from datetime import date, datetime
from enum import StrEnum
from functools import cached_property
from typing import TYPE_CHECKING, Any

import recurrence.fields
from constance import config
from dateutil.relativedelta import relativedelta
from django.conf import settings
from django.contrib.auth.models import AbstractUser, Group
from django.core.cache import cache
from django.core.exceptions import ValidationError
from django.core.validators import MaxValueValidator, MinValueValidator
from django.db import models
from django.db.models.base import ModelBase
from django.db.models.functions.text import Lower
from django.templatetags.static import static
from django.urls import reverse
from django.utils import timezone
from django.utils.translation import gettext as _
from django_stubs_ext.db.models import TypedModelMeta
from encrypted_fields import EncryptedJSONField
from strategy_field.fields import StrategyField
from timezone_field import TimeZoneField
from valkey import Valkey

from birder.checks.base import BaseCheck
from birder.checks.registry import registry
from birder.exceptions import CheckError
from birder.signals import monitor_update
from birder.utils.security import get_random_token
from birder.ws.utils import notify_ui

KEY_ERROR_COUNT = "{0}:monitor:{1.pk}:count"
KEY_STATUS = "{0}:monitor:{1.pk}:status"
KEY_LAST_CHECK = "{0}:monitor:{1.pk}:last_check"
KEY_LAST_SUCCESS = "{0}:monitor:{1.pk}:last_success"
KEY_LAST_FAILURE = "{0}:monitor:{1.pk}:last_failure"
KEY_SYSTEM_CHECK = "{0}:system:last_check"
KEY_PROGRAM_STATUS = "{0}:program:{1.pk}:program_status"

KEY_PROGRAM_CHECKS = "{0}:program:{1.pk}:checks"

valkey = Valkey.from_url(settings.DRAMATIQ_VALKEY_URL)

if TYPE_CHECKING:
    from django.db.models.manager import _T


def get_cache_key(pattern: str, *args: Any) -> str:
    return pattern.format(config.CACHE_PREFIX, *args)


class User(AbstractUser):
    time_zone = TimeZoneField(default="UTC")


class Project(models.Model):
    environments: "models.ManyToManyField[Environment, Environment]"
    name = models.CharField(max_length=255, unique=True)
    public = models.BooleanField(default=False)
    bitcaster_url = models.URLField(blank=True, help_text="The URL to the Bitcaster notification endpoint.")
    environments = models.ManyToManyField("Environment", related_name="projects", blank=False)
    icon = models.CharField(blank=True, default="", max_length=255)
    default_environment = models.ForeignKey("Environment", null=True, on_delete=models.PROTECT)  # type: ignore[misc]

    class Meta:
        ordering = ["name"]
        constraints = [
            models.UniqueConstraint(Lower("name"), name="unique_program_name"),
        ]

    def __str__(self) -> str:
        return self.name

    def save(
        self,
        *,
        force_insert: bool | tuple[ModelBase, ...] = False,
        force_update: bool = False,
        using: str | None = None,
        update_fields: Iterable[str] | None = None,
    ) -> None:
        if self.pk and not self.default_environment:
            self.default_environment = self.environments.first()
        super().save(force_insert=force_insert, force_update=force_update, using=using, update_fields=update_fields)

    def get_absolute_url(self) -> str:
        env = self.default_environment if self.default_environment else self.environments.first()
        return reverse("project-env", kwargs={"project_id": self.pk, "env": env.name})

    def clean(self) -> None:
        if (
            self.pk
            and self.default_environment
            and not self.environments.filter(pk=self.default_environment.pk).exists()
        ):
            raise ValidationError(_("Default environment must be one of selected environment"))
        super().clean()

    @cached_property
    def data(self) -> dict:
        v = valkey.hgetall(get_cache_key(KEY_PROGRAM_CHECKS, self))
        return {k.decode("utf-8"): v.decode("utf-8") for k, v in v.items()}  # type: ignore[union-attr]

    @cached_property
    def failures(self) -> int:
        return len([k for (k, v) in self.data.items() if v == "ko"])

    @cached_property
    def warnings(self) -> int:
        return len([k for (k, v) in self.data.items() if v == "warn"])

    @cached_property
    def success(self) -> int:
        return len([k for (k, v) in self.data.items() if v == "ok"])

    @property
    def status(self) -> str:
        values = self.data.values()
        if Monitor.Status.FAIL in values:
            return Monitor.Status.FAIL
        if Monitor.Status.WARN in values:
            return Monitor.Status.WARN
        if Monitor.Status.SUCCESS in values:
            return Monitor.Status.SUCCESS
        return Monitor.Status.UNKNOWN


class Environment(models.Model):
    name = models.CharField(max_length=255, unique=True)

    class Meta:
        ordering = ["name"]
        constraints = [
            models.UniqueConstraint(Lower("name"), name="unique_env_name"),
        ]

    def __str__(self) -> str:
        return self.name


class UserRole(models.Model):
    user = models.ForeignKey(User, on_delete=models.CASCADE)
    project = models.ForeignKey(Project, on_delete=models.CASCADE)
    role = models.ForeignKey(Group, on_delete=models.CASCADE)

    def __str__(self) -> str:
        return self.user.username


class MonitorQuerySet(models.query.QuerySet):
    pass


class MonitorManager(models.Manager):
    queryset_class = MonitorQuerySet

    def get_queryset(self) -> "models.QuerySet[_T]":
        return super().get_queryset().select_related("environment", "project")


class Monitor(models.Model):
    class Status(StrEnum):
        SUCCESS = "ok"
        WARN = "warn"
        FAIL = "ko"
        UNKNOWN = "question"

    strategy: "BaseCheck"
    project = models.ForeignKey(Project, on_delete=models.CASCADE)
    environment = models.ForeignKey(Environment, on_delete=models.SET_NULL, null=True, blank=False)  # type: ignore[misc]
    name = models.CharField(max_length=255, unique=True)
    position = models.PositiveIntegerField(default=0)
    description = models.TextField(blank=True, help_text="short description  to display in the monitor detail page")
    notes = models.TextField(blank=True, help_text="Notes about the monitor. Only visible to Staff")
    custom_icon = models.CharField(blank=True, default="", max_length=255)
    strategy = StrategyField(registry=registry)
    configuration = EncryptedJSONField(default=dict, help_text="Checker configuration")
    data = models.BinaryField(blank=True, null=True, default=None)  # type: ignore[misc]
    data_file = models.FileField(blank=True, null=True, default=None)

    token = models.CharField(
        default=get_random_token,
        blank=True,
        max_length=1000,
        editable=False,
        help_text="Token to use for external API invocation",
    )

    active = models.BooleanField(default=True)
    warn_threshold = models.PositiveIntegerField(
        default=1,
        validators=[MinValueValidator(0), MaxValueValidator(9)],
        help_text="how many consecutive failures "
        "(or missing notifications in case or remote invocation) produce a warning",
    )
    err_threshold = models.PositiveIntegerField(
        default=5,
        validators=[MinValueValidator(0), MaxValueValidator(9)],
        help_text="how many consecutive failures "
        "(or missing notifications in case or remote invocation) produce an error",
    )

    objects = MonitorManager()

    class Meta(TypedModelMeta):
        ordering = (
            "project__name",
            "name",
        )
        constraints = [
            models.UniqueConstraint("project", Lower("name"), name="unique_project_monitor_name"),
        ]

    def __str__(self) -> str:
        return f"{self.project.name}/{self.name} ({self.environment.name})"

    def get_absolute_url(self) -> str:
        return reverse(
            "monitor-detail", kwargs={"project_id": self.project.pk, "env": self.environment.name, "pk": self.pk}
        )

    @cached_property
    def icon(self) -> str:
        if self.custom_icon and self.custom_icon.startswith("http"):
            return self.custom_icon
        if self.custom_icon:
            return static(f"images/icons/{self.custom_icon}")
        return static(f"images/icons/{self.strategy.icon}")

    def store_error(self, timestamp: datetime) -> None:
        """Set corresponding minute of the day's bit."""
        from .db import DataStore

        ds = DataStore(self)
        ds.store_error(timestamp)

    def reset_current_errors(self) -> int:
        cache.set(get_cache_key(KEY_ERROR_COUNT, self), 0, timeout=86400)
        return 0

    def incr_current_errors(self) -> int:
        try:
            new_value = cache.incr(get_cache_key(KEY_ERROR_COUNT, self), 1)
        except ValueError:
            new_value = 1
            cache.set(get_cache_key(KEY_ERROR_COUNT, self), new_value, timeout=86400)
        return new_value

    def get(self) -> None:
        self.store_last_timestamp_success()
        self.reset_current_errors()
        timestamp = datetime.now()
        self.store_error(timestamp)

    def run(self) -> bool:
        timestamp = datetime.now()
        self.store_last_timestamp_check()
        if self.strategy.mode == BaseCheck.LOCAL_TRIGGER:
            try:
                result = self.strategy.check(raise_error=True)
            except CheckError:
                result = False

            if result:
                self.reset_current_errors()
                self.store_last_timestamp_success()
                st = Monitor.Status.SUCCESS
            else:
                self.store_error(timestamp)
                self.store_last_timestamp_failure()
                error_count = self.incr_current_errors()
                if error_count >= self.err_threshold:
                    st = Monitor.Status.FAIL
                elif error_count >= self.warn_threshold:
                    st = Monitor.Status.WARN
                else:
                    st = Monitor.Status.SUCCESS
            cache.set(get_cache_key(KEY_STATUS, self), st, timeout=86400)

            key = get_cache_key(KEY_PROGRAM_CHECKS, self.project)
            valkey.hset(key, str(self.pk), st)
        else:
            result = False
            if self.last_timestamp_success:
                time_difference = timestamp - self.last_timestamp_success
                offset = int(time_difference.total_seconds() // 60)
                if offset:
                    self.incr_current_errors()
                    self.store_last_timestamp_failure()
                else:
                    self.store_last_timestamp_success()
            else:
                self.incr_current_errors()
                self.store_last_timestamp_failure()

        notify_ui("update", self)
        monitor_update.send(sender=Monitor, instance=self, result=result, timestamp=timestamp)
        return result

    @property
    def counters(self) -> tuple[int, int, int]:
        return self.failures, self.warn_threshold, self.err_threshold

    @property
    def status(self) -> str:
        error_count = self.failures
        if error_count >= self.err_threshold:
            st = Monitor.Status.FAIL
        elif error_count >= self.warn_threshold:
            st = Monitor.Status.WARN
        elif not self.last_timestamp_check:
            st = Monitor.Status.UNKNOWN
        else:
            st = Monitor.Status.SUCCESS
        return st

    @property
    def failures(self) -> int:
        return cache.get(get_cache_key(KEY_ERROR_COUNT, self)) or 0

    def store_last_timestamp_check(self) -> None:
        timestamp = datetime.now()
        ts = timestamp.strftime(config.DATETIME_FORMAT)
        return cache.set(get_cache_key(KEY_LAST_CHECK, self), ts, timeout=86400)

    def store_last_timestamp_failure(self) -> None:
        timestamp = datetime.now()
        ts = timestamp.strftime(config.DATETIME_FORMAT)
        return cache.set(get_cache_key(KEY_LAST_FAILURE, self), ts, timeout=86400)

    def store_last_timestamp_success(self) -> None:
        timestamp = datetime.now()
        ts = timestamp.strftime(config.DATETIME_FORMAT)
        return cache.set(get_cache_key(KEY_LAST_SUCCESS, self), ts, timeout=86400)

    @property
    def last_timestamp_check(self) -> datetime | None:
        try:
            return datetime.strptime(cache.get(get_cache_key(KEY_LAST_CHECK, self)), config.DATETIME_FORMAT)
        except (ValueError, TypeError):
            return None

    @property
    def last_timestamp_success(self) -> datetime | None:
        try:
            return datetime.strptime(cache.get(get_cache_key(KEY_LAST_SUCCESS, self)), config.DATETIME_FORMAT)
        except (ValueError, TypeError):
            return None

    @property
    def last_timestamp_failure(self) -> datetime | None:
        try:
            return datetime.strptime(cache.get(get_cache_key(KEY_LAST_FAILURE, self)), config.DATETIME_FORMAT)
        except (ValueError, TypeError):
            return None

    def regenerate_token(self, save: bool = True) -> None:
        self.token = get_random_token()
        if save:
            self.save()


class DataHistory(models.Model):
    monitor = models.ForeignKey(Monitor, on_delete=models.CASCADE, related_name="datalog")
    date = models.DateField(auto_now_add=False)
    data = models.BinaryField(default=None, null=True)  # type: ignore[misc]

    class Meta:
        ordering = ["-date"]
        constraints = [
            models.UniqueConstraint("monitor", "date", name="unique_data_day_monitor"),
        ]

    def __str__(self) -> str:
        return f"{self.pk}"


class LogCheck(models.Model):
    monitor = models.ForeignKey(Monitor, on_delete=models.CASCADE, related_name="logs")
    status = models.CharField(max_length=255, default="")
    timestamp = models.DateTimeField(default=timezone.now)
    payload = models.TextField(blank=True)

    class Meta:
        ordering = ["-timestamp"]

    def __str__(self) -> str:
        return self.monitor.name


class Deadline(models.Model):
    monitor = models.ForeignKey(Monitor, on_delete=models.CASCADE, related_name="deadlines")
    title = models.CharField(max_length=100)
    description = models.TextField(blank=True)
    start = models.DateField(default=timezone.now)
    end = models.DateField(blank=True, null=True)  # type: ignore[misc]
    recurrences = recurrence.fields.RecurrenceField()
    warn_threshold = models.IntegerField(default=7, help_text="How many days before the deadline should warn monitor")
    alarm_threshold = models.IntegerField(default=1, help_text="How many days before the deadline should alarm monitor")

    class Meta:
        ordering = ["-start"]

    def __str__(self) -> str:
        return self.title

    @cached_property
    def next(self) -> date:
        return self.recurrences.after(datetime.today(), inc=True).date()

    @property
    def is_warn(self) -> bool:
        warn_offset = (datetime.today() - relativedelta(days=self.warn_threshold)).date()
        return self.next > warn_offset

    @property
    def is_expiring(self) -> bool:
        alarm_offset = (datetime.today() - relativedelta(days=self.warn_threshold)).date()
        return self.next > alarm_offset



# Generated by Django 5.2.1 on 2025-05-28 05:40

from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("birder", "0004_project_default_environment_alter_datahistory_date"),
    ]

    operations = [
        migrations.RenameField(
            model_name="monitor",
            old_name="configuration",
            new_name="config_store",
        ),
    ]



import os
from typing import Any

import click

import birder


@click.group()
@click.version_option(version=birder.VERSION, message="Birder %(version)s")
def cli(**kwargs: Any) -> None:
    import django

    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "birder.config.settings")
    django.setup()


def main() -> None:
    cli(prog_name=birder.NAME, obj={}, max_content_width=100)


from . import check, deadlines, monitor, upgrade  # noqa: F401,E402



# Generated by Django 5.2.1 on 2025-05-28 05:47

import encrypted_fields.fields
from django.db import migrations


class Migration(migrations.Migration):
    dependencies = [
        ("birder", "0005_rename_configuration_monitor_config_store"),
    ]

    operations = [
        migrations.RemoveField(
            model_name="monitor",
            name="config_store",
        ),
        migrations.AddField(
            model_name="monitor",
            name="configuration",
            field=encrypted_fields.fields.EncryptedJSONField(default=dict, help_text="Checker configuration"),
        ),
    ]



from textwrap import shorten
from typing import Any

import click

from . import cli

KEY = "birder:upgrade"


@cli.group()
def deadlines(**kwargs: Any) -> None:
    """Manage monitors."""


@deadlines.command(name="list")
def list_(**kwargs: Any) -> None:  # noqa: C901, PLR0912
    """Upgrade system."""
    from birder.models import Deadline

    for deadline in Deadline.objects.all():
        if deadline.is_expiring:
            color = "red"
        elif deadline.is_warn:
            color = "yellow"
        else:
            color = "green"

        click.secho(
            f"{deadline.next} "
            f"{shorten(deadline.monitor.project.name, width=20):<22}"
            f"{shorten(deadline.monitor.name, width=20):<22}"
            f"{shorten(deadline.monitor.environment.name, width=20):<22}"
            f"{deadline.title} ",
            fg=color,
        )



from django.utils.translation import gettext_lazy
from unfold.sites import UnfoldAdminSite


class ManageSite(UnfoldAdminSite):
    site_title = gettext_lazy("Birder site admin")
    site_header = gettext_lazy("Birder")
    index_title = gettext_lazy("Birder")
    site_url = "/"
    index_template = None
    settings_name = "MANAGE_CONFIG"



# Generated by Django 5.2.2 on 2025-06-16 10:30

import django.db.models.deletion
import django.utils.timezone
import recurrence.fields
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("birder", "0006_remove_monitor_config_store_monitor_configuration"),
    ]

    operations = [
        migrations.CreateModel(
            name="Deadline",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("title", models.CharField(max_length=100)),
                ("description", models.TextField(blank=True)),
                ("start", models.DateField(default=django.utils.timezone.now)),
                ("end", models.DateField(blank=True, null=True)),
                ("recurrences", recurrence.fields.RecurrenceField()),
                (
                    "warn_threshold",
                    models.IntegerField(default=7, help_text="How many days before the deadline should warn monitor"),
                ),
                (
                    "alarm_threshold",
                    models.IntegerField(default=1, help_text="How many days before the deadline should alarm monitor"),
                ),
                (
                    "monitor",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, related_name="deadlines", to="birder.monitor"
                    ),
                ),
            ],
            options={
                "ordering": ["-start"],
            },
        ),
    ]



import boto3
from django import forms

from ..exceptions import CheckError
from . import HttpCheck
from .base import ConfigForm, WriteOnlyField


class S3Config(ConfigForm):
    endpoint_url = forms.CharField(required=True)
    bucket_name = forms.CharField(required=True)
    region_name = forms.CharField(required=False)
    aws_access_key_id = forms.CharField(required=True)
    aws_secret_access_key = WriteOnlyField(required=True)
    aws_session_token = WriteOnlyField(required=False)


class S3Check(HttpCheck):
    icon = "s3.svg"
    pragma = ["s3"]
    config_class = S3Config
    address_format: str = "s3://{bucket_name}/"

    def check(self, raise_error: bool = False) -> bool:
        try:
            cfg = {**self.config}
            bucket_name = cfg.pop("bucket_name")
            s3 = boto3.resource("s3", **cfg, config=boto3.session.Config(signature_version="s3v4"), verify=False)
            s3.Bucket(bucket_name).check()
            return True
        except ValueError as e:
            if raise_error:
                raise CheckError("JSON check failed") from e
        return False



import asyncio
from typing import Any

import amqp.exceptions
import kombu.exceptions
import redis.exceptions
from celery import Celery as CeleryApp
from celery.app.control import Control
from celery.exceptions import CeleryError
from django import forms
from django.core.validators import MaxValueValidator, MinValueValidator
from flower.utils.broker import Broker

from ..exceptions import CheckError
from .base import BaseCheck, ConfigForm


class CeleryConfig(ConfigForm):
    broker = forms.ChoiceField(choices=[("amqp", "amqp"), ("redis", "redis")])
    hostname = forms.CharField()
    port = forms.IntegerField(required=True)
    extra = forms.CharField(required=False)
    min_workers = forms.IntegerField(
        required=True, validators=[MinValueValidator(1)], help_text="Minimum number of workers", initial=1
    )
    timeout = forms.IntegerField(validators=[MinValueValidator(1), MaxValueValidator(5)], initial=2)


class CeleryCheck(BaseCheck):
    icon = "celery.svg"
    pragma = ["celery"]
    config_class = CeleryConfig
    address_format = "{broker}://{hostname}:{port}/{extra}"

    @classmethod
    def clean_config(cls, cfg: dict[str, Any]) -> dict[str, Any]:
        if not cfg.get("hostname"):
            cfg["hostname"] = cfg.get("host", "")
        return cfg

    def check(self, raise_error: bool = False) -> bool:
        try:
            broker = "{broker}://{hostname}:{port}/{extra}".format(**self.config)
            app = CeleryApp("birder", loglevel="info", broker=broker)
            ctrl = Control(app)
            workers = len(ctrl.ping())
            self.status = {"workers": workers}
            return workers > self.config["min_workers"]
        except (
            CeleryError,
            KeyError,
            redis.exceptions.RedisError,
            kombu.exceptions.KombuError,
            amqp.exceptions.AMQPError,
        ) as e:
            if raise_error:
                raise CheckError("Celery check failed") from e
        return False


class CeleryQueueConfig(ConfigForm):
    broker = forms.ChoiceField(choices=[("amqp", "amqp"), ("redis", "redis")])
    hostname = forms.CharField()
    port = forms.IntegerField(required=True)
    extra = forms.CharField(required=False)
    queue_name = forms.CharField(required=True, initial="celery")

    max_queued = forms.IntegerField(
        required=True, validators=[MinValueValidator(1)], initial=1, help_text="Max number of elements pending queue"
    )
    timeout = forms.IntegerField(validators=[MinValueValidator(1), MaxValueValidator(5)], initial=2)


class CeleryQueueCheck(BaseCheck):
    icon = "celery.svg"
    pragma = ["celery+queue"]
    config_class = CeleryQueueConfig
    address_format = "{broker}://{hostname}:{port}/{extra}"

    @classmethod
    def clean_config(cls, cfg: dict[str, Any]) -> dict[str, Any]:
        if not cfg.get("hostname"):
            cfg["hostname"] = cfg.get("host", "")
        if not cfg.get("min_workers"):
            cfg["min_workers"] = 1
        return cfg

    def check(self, raise_error: bool = False) -> bool:
        try:
            broker = Broker("{broker}://{hostname}:{port}/{extra}".format(**self.config))
            queues_result = broker.queues([self.config["queue_name"]])
            res = asyncio.run(queues_result) or [{"messages": 0}]
            length = res[0].get("messages", 0)
            self.status = {"size": length}
            return length > self.config["max_queued"]
        except (
            CeleryError,
            KeyError,
            NotImplementedError,
            redis.exceptions.RedisError,
            kombu.exceptions.KombuError,
            amqp.exceptions.AMQPError,
        ) as e:
            if raise_error:
                raise CheckError("Celery check failed") from e
        return False



from functools import cached_property
from typing import Any

from .base import BaseCheck, ConfigForm


class HealthCheckConfig(ConfigForm):
    help_text = """
{% absolute_url "trigger" monitor.pk monitor.token %}
"""

    def is_valid(self) -> bool:
        self.cleaned_data = {}
        return True


class HealthCheck(BaseCheck):
    icon = "socket.svg"
    pragma = []
    config_class = HealthCheckConfig
    address_format = ""
    mode = BaseCheck.REMOTE_INVOCATION
    verbose_name = "Remote HealthCheck"

    @classmethod
    def clean_config(cls, cfg: dict[str, Any]) -> dict[str, Any]:
        return {}

    @cached_property
    def config(self) -> dict[str, Any]:
        return self.config_class.DEFAULTS

    @property
    def address(self) -> str:
        return "-"

    def check(self, raise_error: bool = False) -> bool:
        return True



import socket
from contextlib import closing

from django import forms
from django.core.validators import MaxValueValidator, MinValueValidator

from ..exceptions import CheckError
from .base import BaseCheck, ConfigForm


class TCPConfig(ConfigForm):
    hostname = forms.CharField(required=True, help_text="Hostname or IP Address")
    port = forms.IntegerField(validators=[MinValueValidator(1)], initial=7)
    timeout = forms.IntegerField(validators=[MinValueValidator(1), MaxValueValidator(5)], initial=2)


class TCPCheck(BaseCheck):
    icon = "tcp.svg"
    pragma = ["tcp"]
    config_class = TCPConfig
    address_format = "{hostname}:{port}"

    def check(self, raise_error: bool = False) -> bool:
        try:
            with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
                socket.setdefaulttimeout(self.config["timeout"])  # seconds (float)
                result = sock.connect_ex((self.config["hostname"], self.config["port"]))
                if result != 0:
                    raise CheckError("TCP Server timed out")
                return True
        except (EOFError, TimeoutError, ConnectionRefusedError, CheckError) as e:
            if raise_error:
                raise CheckError("TCP check failed") from e
        return False



# Generated by Django 5.2.1 on 2025-05-26 09:10

import django.db.models.deletion
from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("birder", "0003_remove_monitor_verbosity"),
    ]

    operations = [
        migrations.AddField(
            model_name="project",
            name="default_environment",
            field=models.ForeignKey(null=True, on_delete=django.db.models.deletion.PROTECT, to="birder.environment"),
        ),
        migrations.AlterField(
            model_name="datahistory",
            name="date",
            field=models.DateField(),
        ),
    ]



from typing import Any

import click
from click import Context
from tabulate import tabulate

from . import cli


@cli.group()
def monitor(**kwargs: Any) -> None:
    """Manage monitors."""


@monitor.command(name="list")
@click.pass_context
def list_(ctx: Context, **kwargs: Any) -> None:
    """List all existing monitors."""
    from birder.models import Monitor

    data = Monitor.objects.values(
        "id",
        "project__name",
        "name",
        "strategy",
        "active",
    )
    table = tabulate(data, [], tablefmt="simple")
    click.echo(table)


@monitor.command(name="check")
@click.argument("monitor_id", type=int, required=False)
@click.option("-a", "--all", "_all", type=int, is_flag=True)
@click.option("-d", "--debug", "debug", is_flag=True)
@click.pass_context
def check_(ctx: Context, monitor_id: int, _all: bool = False, debug: bool = True, **kwargs: Any) -> None:
    """Run selected check."""
    from birder.models import BaseCheck, Monitor

    if _all and monitor_id:
        raise click.UsageError("Do not use --all and id")
    ok = click.style("\u2714", fg="green")
    ko = click.style("\u2716", fg="red")

    def c(m: Monitor) -> None:
        res = m.run()
        status = ok if res else ko
        info = "" if not debug else f" | {m.configuration}"
        click.echo(
            f"{m.project.name[:20]:<22} | "
            f"{m.environment.name[:15]:<17} | "
            f"{m.name[:20]:<22} | "
            f"{status} | "
            f"{m.strategy.status}"
            f"{info}"
        )

    if _all:
        monitors = Monitor.objects.select_related("project", "environment").order_by(
            "project__name", "environment", "name"
        )
    else:
        monitors = [Monitor.objects.get(id=monitor_id)]

    for monitor in monitors:
        if monitor.strategy.mode == BaseCheck.LOCAL_TRIGGER:
            c(monitor)


@monitor.command()
@click.argument("monitor_id", type=int)
@click.pass_context
def refresh(ctx: Context, monitor_id: int, **kwargs: Any) -> None:
    """Force UI refresh."""
    from birder.models import Monitor
    from birder.ws.utils import notify_ui

    monitor = Monitor.objects.get(id=monitor_id)
    notify_ui("update", monitor=monitor)


@monitor.command()
@click.pass_context
def reset(ctx: Context, **kwargs: Any) -> None:
    """Reset all checks."""
    from django.core.cache import cache

    cache.clear()



import logging
from typing import Any

from constance import config
from django.core.exceptions import ValidationError
from django.forms import CharField, ChoiceField, HiddenInput, TextInput, Textarea
from django.template import Context, Template
from django.utils.safestring import SafeString, mark_safe
from durations_nlp.duration import Duration, Scale
from durations_nlp.exceptions import InvalidTokenError, ScaleFormatError

logger = logging.getLogger(__name__)


class ObfuscatedInput(HiddenInput):
    def render(
        self,
        name: str,
        value: Any,
        attrs: dict[str, str] | None = None,
        renderer: Any | None = None,
    ) -> "SafeString":
        context = self.get_context(name, value, attrs)
        context["value"] = str(value)
        context["label"] = "Set" if value else "Not Set"

        tpl = Template('<input type="hidden" name="{{ widget.name }}" value="{{ value }}">{{ label }}')
        return mark_safe(tpl.render(Context(context)))  # noqa: S308


class WriteOnlyWidget:
    def format_value(self, value: Any) -> str:
        return super().format_value("***")

    def value_from_datadict(self, data: dict[str, Any], files: Any, name: str) -> Any:
        value = data.get(name)
        if value == "***":
            return getattr(config, name)
        return value


class WriteOnlyTextarea(WriteOnlyWidget, Textarea):
    pass


class WriteOnlyInput(WriteOnlyWidget, TextInput):
    pass


def format_duration(d: Duration) -> str:
    entry: tuple[float, Scale]
    ret = []
    for entry in d.parsed_durations:
        num = int(entry[0])
        a = entry[1].representation.long_singular
        if num > 1:
            a = entry[1].representation.long_plural
        ret.append(f"{num} {a}")

    return " ".join(ret)


class DurationField(CharField):
    def clean(self, value: str) -> str:
        try:
            d: Duration = Duration(value)
            if d.to_minutes() < 5:  # noqa: PLR2004
                raise ValidationError("Must be more than 5 minutes")
        except (InvalidTokenError, ValidationError, ScaleFormatError) as e:
            raise ValidationError("Invalid duration") from e
        return format_duration(d)


class GroupChoiceField(ChoiceField):
    def __init__(self, **kwargs: Any) -> None:
        from django.contrib.auth.models import Group

        ret: list[tuple[str | int, str]] = [(c["name"], c["name"]) for c in Group.objects.values("pk", "name")]
        kwargs["choices"] = ret
        super().__init__(**kwargs)



from typing import Any

from django.core.cache import cache
from django.http.request import HttpRequest

from birder import VERSION


def birder(request: HttpRequest) -> dict[str, Any]:
    return {
        "birder": {
            "system": {
                "last_check": cache.get("system:last_check"),
            },
            "version": VERSION,
        }
    }



from typing import Any

from django import forms
from django.core.validators import MaxValueValidator, MinValueValidator
from ldap3 import SYNC, Connection, Server
from ldap3.core.exceptions import LDAPException

from ..exceptions import CheckError
from .base import BaseCheck, ConfigForm, WriteOnlyField


class LDAPConfig(ConfigForm):
    host = forms.CharField(required=True, help_text="Server address")
    port = forms.IntegerField(validators=[MinValueValidator(1)], initial=389)
    version = forms.IntegerField(validators=[MinValueValidator(1), MaxValueValidator(3)], initial=3)

    user = forms.CharField(required=True)
    password = WriteOnlyField(required=True)


class LDAPCheck(BaseCheck):
    icon = "ldap.svg"
    pragma = ["ldap"]
    config_class = LDAPConfig
    address_format = "{host}:{port}"
    client_strategy = SYNC

    @classmethod
    def clean_config(cls, cfg: dict[str, Any]) -> dict[str, Any]:
        if not cfg.get("password"):
            cfg["password"] = cfg.get("password", "")
        if not cfg.get("user"):
            cfg["user"] = cfg.get("username", "")
        return cfg

    def check(self, raise_error: bool = False) -> bool:
        try:
            config = {**self.config}
            host, port = config.pop("host"), config.pop("port")
            uri = f"ldap://{host}:{port}"
            server = Server(uri)
            cfg = {**config, "client_strategy": self.client_strategy}
            conn = Connection(server, **cfg)
            conn.bind()
            return True
        except LDAPException as e:
            if raise_error:
                raise CheckError(f"LDAP check failed: {e}") from e
        return False



import json
from datetime import datetime, timedelta
from typing import Any

from django.contrib.auth.views import LoginView as LoginView_
from django.db.models import QuerySet
from django.forms import Media
from django.http.request import HttpRequest
from django.http.response import HttpResponse, HttpResponseRedirect
from django.shortcuts import get_object_or_404, render
from django.urls import reverse
from django.utils.safestring import mark_safe
from django.views.generic import DetailView, TemplateView
from django.views.generic.base import ContextMixin, View

from birder.checks import BaseCheck
from birder.config import settings
from birder.forms import LoginForm
from birder.models import Monitor, Project
from birder.utils.dates import format_minutes_as_time, get_start_of_day
from birder.ws.utils import notify_ui


class CommonContextMixin(ContextMixin, View):
    js_files = []

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        kwargs["active_view"] = self.__class__.__name__
        kwargs["sso_enabled"] = bool(settings.SOCIAL_AUTH_GOOGLE_OAUTH2_KEY)
        kwargs["media"] = self.media
        return super().get_context_data(**kwargs)

    @property
    def media(self) -> Media:
        extra = "" if settings.DEBUG else ".min"

        js_files = [
            f"admin/js/vendor/jquery/jquery{extra}.js",
            "admin/js/jquery.init.js",
            f"birder{extra}.js",
            *[f % extra for f in self.js_files],
        ]
        return Media(js=js_files)


class IndexView(CommonContextMixin, TemplateView):
    template_name = "index.html"
    js_files = ["index%s.js"]

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        kwargs["projects"] = Project.objects.filter(public=True)
        return super().get_context_data(**kwargs)


class ProjectRouterView(CommonContextMixin, DetailView):
    model = Project

    def get(self, request: HttpRequest, *args: Any, **kwargs: Any) -> HttpResponse:
        return HttpResponseRedirect(self.get_object().get_absolute_url())


class ProjectView(CommonContextMixin, DetailView):
    template_name = "project.html"
    model = Project
    js_files = ["project%s.js"]

    def get_queryset(self) -> QuerySet[Project]:
        return super().get_queryset().select_related("default_environment")

    def get_object(self, queryset: QuerySet[Project] = None) -> Project:
        return self.get_queryset().get(pk=self.kwargs.get("project_id"))

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        project = self.get_object()
        env = project.environments.get(name=self.kwargs.get("env"))

        kwargs["selected_env"] = env
        filters = {"environment": env}
        kwargs["project"] = project
        monitors = Monitor.objects.filter(**filters).order_by("position", "name")
        kwargs["monitors"] = monitors
        kwargs["environments"] = project.environments.order_by("name")
        for monitor in monitors:
            notify_ui("update", monitor=monitor)
        return super().get_context_data(**kwargs)


class MonitorDetail(CommonContextMixin, DetailView):
    template_name = "monitor.html"
    queryset = Monitor.objects.all()
    js_files = ["https://cdnjs.cloudflare.com/ajax/libs/Chart.js/3.9.1/chart%s.js", "monitor%s.js"]

    def get_queryset(self) -> QuerySet[Monitor]:
        return super().get_queryset().select_related("environment", "project")

    def get_absolute_url(self) -> str:
        return reverse(
            "monitor-detail",
            kwargs={"pk": self.object.pk, "env": self.object.environment.name, "project_id": self.object.project.pk},
        )

    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        from birder.db import DataStore

        now = get_start_of_day(datetime.now())
        bits = DataStore(self.object).get_all_entries(now)
        group_by = 5
        data = [sum(bits[i : i + group_by]) for i in range(0, len(bits), group_by)]
        kwargs["title"] = "{} - {}".format(
            now.strftime("%H:%M"), (now + timedelta(hours=(len(bits) // 60) - 1)).strftime("%H:%M")
        )
        kwargs["data"] = data
        kwargs["group_by"] = group_by
        kwargs["labels"] = mark_safe(  # noqa: S308
            json.dumps([format_minutes_as_time(i) for i in list(range(1, len(bits) + 1, group_by))])
        )
        return super().get_context_data(**kwargs)


class LoginView(CommonContextMixin, LoginView_):
    template_name = "login.html"
    form_class = LoginForm


def trigger(request: HttpRequest, pk: str, token: str) -> HttpResponse:
    m: Monitor = get_object_or_404(Monitor, pk=pk)
    if m.token != token:
        return HttpResponse("Invalid Token", status=403)
    if m.strategy.mode != BaseCheck.REMOTE_INVOCATION:
        return HttpResponse("Check not enabled for remote call", status=400)
    m.get()
    return HttpResponse("Ok")


def error_400(request: HttpRequest, exception: Exception = None) -> HttpResponse:
    return render(request, "errors/400.html", {"error_code": 400, "message": "Bad Request"}, status=400)


def error_403(request: HttpRequest, exception: Exception = None) -> HttpResponse:
    return render(request, "errors/403.html", {"error_code": 403, "message": "Forbidden"}, status=403)


def error_404(request: HttpRequest, exception: Exception = None) -> HttpResponse:
    return render(request, "errors/404.html", {"error_code": 404, "message": "Page not found"}, status=404)


def error_500(request: HttpRequest, exception: Exception = None) -> HttpResponse:
    return render(request, "errors/500.html", {"error_code": 500, "message": "Server Error"}, status=500)



# Generated by Django 5.1.4 on 2025-01-10 17:34

import django.contrib.auth.models
import django.contrib.auth.validators
import django.core.validators
import django.db.models.deletion
import django.db.models.functions.text
import django.utils.timezone
import strategy_field.fields
import timezone_field.fields
from django.conf import settings
from django.db import migrations, models

import birder.utils.security


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("auth", "0012_alter_user_first_name_max_length"),
    ]

    operations = [
        migrations.CreateModel(
            name="User",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("password", models.CharField(max_length=128, verbose_name="password")),
                ("last_login", models.DateTimeField(blank=True, null=True, verbose_name="last login")),
                (
                    "is_superuser",
                    models.BooleanField(
                        default=False,
                        help_text="Designates that this user has all permissions without explicitly assigning them.",
                        verbose_name="superuser status",
                    ),
                ),
                (
                    "username",
                    models.CharField(
                        error_messages={"unique": "A user with that username already exists."},
                        help_text="Required. 150 characters or fewer. Letters, digits and @/./+/-/_ only.",
                        max_length=150,
                        unique=True,
                        validators=[django.contrib.auth.validators.UnicodeUsernameValidator()],
                        verbose_name="username",
                    ),
                ),
                ("first_name", models.CharField(blank=True, max_length=150, verbose_name="first name")),
                ("last_name", models.CharField(blank=True, max_length=150, verbose_name="last name")),
                ("email", models.EmailField(blank=True, max_length=254, verbose_name="email address")),
                (
                    "is_staff",
                    models.BooleanField(
                        default=False,
                        help_text="Designates whether the user can log into this admin site.",
                        verbose_name="staff status",
                    ),
                ),
                (
                    "is_active",
                    models.BooleanField(
                        default=True,
                        help_text="Designates whether this user should be treated as active. Unselect this instead of deleting accounts.",
                        verbose_name="active",
                    ),
                ),
                ("date_joined", models.DateTimeField(default=django.utils.timezone.now, verbose_name="date joined")),
                ("time_zone", timezone_field.fields.TimeZoneField(default="UTC")),
                (
                    "groups",
                    models.ManyToManyField(
                        blank=True,
                        help_text="The groups this user belongs to. A user will get all permissions granted to each of their groups.",
                        related_name="user_set",
                        related_query_name="user",
                        to="auth.group",
                        verbose_name="groups",
                    ),
                ),
                (
                    "user_permissions",
                    models.ManyToManyField(
                        blank=True,
                        help_text="Specific permissions for this user.",
                        related_name="user_set",
                        related_query_name="user",
                        to="auth.permission",
                        verbose_name="user permissions",
                    ),
                ),
            ],
            options={
                "verbose_name": "user",
                "verbose_name_plural": "users",
                "abstract": False,
            },
            managers=[
                ("objects", django.contrib.auth.models.UserManager()),
            ],
        ),
        migrations.CreateModel(
            name="Environment",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=255, unique=True)),
            ],
            options={
                "constraints": [
                    models.UniqueConstraint(django.db.models.functions.text.Lower("name"), name="unique_env_name")
                ],
            },
        ),
        migrations.CreateModel(
            name="Monitor",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=255, unique=True)),
                ("position", models.PositiveIntegerField(default=0)),
                (
                    "description",
                    models.TextField(blank=True, help_text="short description  to display in the monitor detail page"),
                ),
                ("notes", models.TextField(blank=True, help_text="Notes about the monitor. Only visible to Staff")),
                ("custom_icon", models.CharField(blank=True, default="", max_length=255)),
                ("strategy", strategy_field.fields.StrategyField()),
                ("configuration", models.JSONField(default=dict, help_text="Checker configuration")),
                ("data", models.BinaryField(blank=True, default=None, null=True)),
                ("data_file", models.FileField(blank=True, default=None, null=True, upload_to="")),
                (
                    "token",
                    models.CharField(
                        blank=True,
                        default=birder.utils.security.get_random_token,
                        editable=False,
                        help_text="Token to use for external API invocation",
                        max_length=1000,
                    ),
                ),
                ("active", models.BooleanField(default=True)),
                (
                    "warn_threshold",
                    models.PositiveIntegerField(
                        default=1,
                        help_text="how many consecutive failures (or missing notifications in case or remote invocation) produce a warning",
                        validators=[
                            django.core.validators.MinValueValidator(0),
                            django.core.validators.MaxValueValidator(9),
                        ],
                    ),
                ),
                (
                    "err_threshold",
                    models.PositiveIntegerField(
                        default=5,
                        help_text="how many consecutive failures (or missing notifications in case or remote invocation) produce an error",
                        validators=[
                            django.core.validators.MinValueValidator(0),
                            django.core.validators.MaxValueValidator(9),
                        ],
                    ),
                ),
                (
                    "verbosity",
                    models.IntegerField(
                        choices=[(0, "None"), (1, "Success"), (2, "Failure"), (3, "Error"), (4, "Full")], default=0
                    ),
                ),
                (
                    "env",
                    models.ForeignKey(
                        blank=True, null=True, on_delete=django.db.models.deletion.SET_NULL, to="birder.environment"
                    ),
                ),
            ],
            options={
                "ordering": ("name",),
            },
        ),
        migrations.CreateModel(
            name="LogCheck",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("status", models.CharField(default="", max_length=255)),
                ("timestamp", models.DateTimeField(default=django.utils.timezone.now)),
                ("payload", models.TextField(blank=True)),
                (
                    "monitor",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, related_name="logs", to="birder.monitor"
                    ),
                ),
            ],
            options={
                "ordering": ["-timestamp"],
            },
        ),
        migrations.CreateModel(
            name="Project",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("name", models.CharField(max_length=255, unique=True)),
                ("public", models.BooleanField(default=False)),
                ("default", models.BooleanField(default=False)),
                (
                    "bitcaster_url",
                    models.URLField(blank=True, help_text="The URL to the Bitcaster notification endpoint."),
                ),
            ],
            options={
                "constraints": [
                    models.UniqueConstraint(django.db.models.functions.text.Lower("name"), name="unique_program_name")
                ],
            },
        ),
        migrations.AddField(
            model_name="monitor",
            name="project",
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="birder.project"),
        ),
        migrations.CreateModel(
            name="UserRole",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("project", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="birder.project")),
                ("role", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to="auth.group")),
                ("user", models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.CreateModel(
            name="DataHistory",
            fields=[
                ("id", models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name="ID")),
                ("date", models.DateField(auto_now_add=True)),
                ("data", models.BinaryField(default=None, null=True)),
                (
                    "monitor",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE, related_name="datalog", to="birder.monitor"
                    ),
                ),
            ],
            options={
                "constraints": [
                    models.UniqueConstraint(models.F("monitor"), models.F("date"), name="unique_data_day_monitor")
                ],
            },
        ),
        migrations.AddConstraint(
            model_name="monitor",
            constraint=models.UniqueConstraint(
                models.F("project"), django.db.models.functions.text.Lower("name"), name="unique_project_monitor_name"
            ),
        ),
    ]



from datetime import datetime, timedelta

from bitarray._bitarray import bitarray
from django.core.cache import cache

from birder.models import DataHistory, Monitor


def init_bitaarray(initial_value: int = 0) -> bitarray:
    bt = bitarray(1440)
    bt.setall(initial_value)
    return bt


class DataStore:
    def __init__(self, monitor: Monitor, prefix: str = "ds") -> None:
        self.monitor: Monitor = monitor
        self.prefix: str = prefix

    def get_all_entries(self, timestamp: datetime) -> bitarray:
        return self._get_store_for_date(timestamp.strftime("%Y-%m-%d"))

    def get_hour_entries(self, timestamp: datetime, hour: int = None) -> bitarray:
        data = self._get_store_for_date(timestamp.strftime("%Y-%m-%d"))
        if not hour:
            hour = timestamp.hour
        start_index = hour * 60
        end_index = start_index + 60
        return data[start_index:end_index]

    def _get_store_for_date(self, sig: str) -> bitarray:
        stored = cache.get(f"{self.prefix}:monitor:{self.monitor.pk}:data:{sig}")
        if stored:
            data = bitarray()
            data.frombytes(stored)
        else:
            data = init_bitaarray(0)
        return data

    def store_error(self, timestamp: datetime) -> int:
        sig = timestamp.strftime("%Y-%m-%d")
        data = self._get_store_for_date(sig)
        absolute_minute = timestamp.hour * 60 + timestamp.minute
        data[absolute_minute] = 1
        cache.set(f"{self.prefix}:monitor:{self.monitor.pk}:data:{sig}", data.tobytes(), timeout=86400)
        return absolute_minute

    def archive(self, timestamp: datetime) -> None:
        for i in range(1, 8):
            day = timestamp - timedelta(days=i)
            sig = day.strftime("%Y-%m-%d")
            stored = self._get_store_for_date(sig)
            if stored:
                DataHistory.objects.update_or_create(monitor=self.monitor, date=day, defaults={"data": stored})
                cache.delete(f"{self.prefix}:monitor:{self.monitor.pk}:data:{sig}")



from functools import cached_property
from typing import TYPE_CHECKING, Any

from django import forms
from django.forms.forms import DeclarativeFieldsMetaclass
from django.http import HttpRequest
from django.template.base import Template
from django.template.context import Context
from django.utils.functional import SimpleLazyObject
from django.utils.safestring import mark_safe
from markdown_deux import markdown
from unfold.widgets import BASE_INPUT_CLASSES

if TYPE_CHECKING:
    from birder.models import Monitor
    from birder.types import Json


class DefaultsMetaclass(DeclarativeFieldsMetaclass):
    def __new__(cls, name: str, bases: tuple[type, ...], attrs: Any) -> type:
        new_class = super().__new__(cls, name, bases, attrs)
        initial_values = {}
        # if not hasattr(new_class, "DEFAULTS"):
        for field_name, field in new_class.declared_fields.items():
            initial_values[field_name] = field.initial
        new_class.DEFAULTS = initial_values
        return new_class


class WriteOnlyWidget(forms.TextInput):
    def __init__(self, attrs: dict[str, Any] | None = None, render_value: bool = False) -> None:
        super().__init__(attrs)
        self.render_value = render_value

    def get_context(self, name: str, value: Any, attrs: Any) -> dict[str, Any]:
        if not self.render_value:
            value = WriteOnlyField.MASK
        return super().get_context(name, value, attrs)


class WriteOnlyField(forms.CharField):
    widget = WriteOnlyWidget
    MASK = "***"


class ConfigForm(forms.Form, metaclass=DefaultsMetaclass):
    help_text = ""

    def __init__(
        self,
        request: HttpRequest | None = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        super().__init__(request, *args, **kwargs)
        for f in self.fields.values():
            f.widget.attrs["class"] = " ".join(BASE_INPUT_CLASSES)

    def render_help(self, context: dict[str, Any], **extra_context: Any) -> str:
        return mark_safe(  # noqa: S308
            markdown(
                Template("{%% load birder %%}%s" % self.help_text)  # noqa: UP031
                .render(Context({**context, **extra_context}))
                .strip()
            )
        )

    @property
    def media(self) -> forms.Media:
        media = super().media
        media += forms.Media(
            js=[
                "admin/js/vendor/jquery/jquery.js",
                "admin/js/jquery.init.js",
                "change-icon.js",
            ],
            css={"screen": ["birder-admin.css"]},
        )
        return media

    @cached_property
    def changed_data(self) -> list[str]:
        return [
            name for name, bf in self._bound_items() if not isinstance(bf.field, WriteOnlyField) and bf._has_changed()
        ] + [
            name
            for name, bf in self._bound_items()
            if isinstance(bf.field, WriteOnlyField) and self.cleaned_data[name] != self.initial[name]
        ]

    def full_clean(self) -> None:
        super().full_clean()
        if self.is_bound:  # Stop further processing.
            for k, v in self.cleaned_data.items():
                if isinstance(self.fields[k], WriteOnlyField) and v == WriteOnlyField.MASK:
                    self.cleaned_data[k] = self.initial.get(k)


class BaseCheck:
    LOCAL_TRIGGER = 1
    REMOTE_INVOCATION = 2

    icon: str
    pragma: list[str]
    config_class: type[ConfigForm]
    mode = LOCAL_TRIGGER
    address_format: str = ""
    verbose_name = None

    def __init__(self, owner: "Monitor|None" = None, configuration: "Json | None" = None) -> None:
        if owner:
            self._configuration = SimpleLazyObject(lambda: owner.configuration)
        elif configuration is not None:
            self._configuration = SimpleLazyObject(lambda: configuration)
        else:
            raise ValueError("Must specify a configuration")  # pragma: no cover
        self.monitor: Monitor = owner
        self.status = {}

    @classmethod
    def clean_config(cls, cfg: dict[str, Any]) -> dict[str, Any]:
        return cfg

    @cached_property
    def config(self) -> dict[str, Any]:
        cfg = {**self.config_class.DEFAULTS, **dict(self._configuration)}
        frm = self.config_class(cfg)
        if frm.is_valid():
            return frm.cleaned_data
        return self.config_class.DEFAULTS

    @property
    def address(self) -> str:
        return self.address_format.format(**self.config)

    def check(self, raise_error: bool = False) -> bool:
        """Perform the check."""
